import pandas as pd
import numpy as np
import copy as cp
from . import hmm_utilities
import time
from intervalframe import IntervalFrame


def get_seq_info(genomeBuild = "hg19",
                 genomeStyle = "NCBI"):
    """
    """

    # Read Chromosome lengths
    seq_info = {}
    for line in open(genomeBuild+"_chrom_sizes.txt", "r"):
        fields = line.strip().split("\t")

        # Manage chr's
        if genomeStyle == "NCBI":
            seq_info[fields[0][3:]] = int(fields[1])
        else:
            seq_info[fields[0]] = int(fields[1])

    return seq_info

def wigToSeries(wig_fn):
    """
    """

    # Set up recordings
    bin_value = defaultdict(list)
    bin_index = defaultdict(list)

    # Read wig file
    for line in open(wig_fn, "r"):
        fields = line.strip().split(" ")
        
        # Check for header
        if len(fields) > 1:
            chrom = fields[1].split("=")[1]
            start = int(fields[2].split("=")[1])
            end = int(fields[3].split("=")[1])
            bin_size = int(fields[4].split("=")[1])
        # Record bin value
        else:
            value = int(fields[0])
            bin_value[chrom].append(value)
            bin_index[chrom].append(start)
            start += bin_size

    # Convert lists to Series
    wig_results = {}
    for chrom in bin_value:
        wig_results = pd.Series(np.array(bin_value[chrom]),
                                index=bin_index[chrom])

    return wig_results, bin_size


def hmmCNV(tumour_copy,
           normal = [0.2, 0.5, 0.75],
           ploidy = [1, 2, 3],
           scStates = None,
           lambda_p = None,
           lambdaScaleHyperParam = 3,
           estimateNormal = True,
           estimatePloidy = True,
           estimateScPrevalence = True,
           maxFracCNASubclone = 0.7,
           maxFracGenomeSubclone = 0.5,
           minSegmentBins = 50,
           altFracThreshold = 0.05,
           coverage = None,
           maxCN = 7,
           txnE = 0.9999999,
           txnStrength = 1e7,
           normalizeMaleX = True,
           includeHOMD = False,
           fracReadsInChrYForMale = 0.001,
           chrXMedianForMale = -0.1,
           outDir = "./",
           libdir = None,
           plotFileType = "pdf",
           plotYLim = (-2,2),
           gender = None,
           genomeBuild = "hg19",
           genomeStyle = "UCSC",
           chrs = np.append(np.arange(1,23).astype(str), "X"),
           chrTrain = np.arange(1, 23).astype(str),
           chrNormalize = np.arange(1, 23).astype(str),
           verbose=False):
    """
    Run ichor HMM copy algorithm

    Parameters
    ----------
        tumour_copy : IntervalFrame

        normal : list of int

        ploidy : list of int

        scStates : 

        lambda_p : float

        lambdaScaleHyperParam : int

        estimateNormal : bool

        estimatePloidy : bool

        estimateScPrevalence : bool

        maxFracCNASubclone : float

        maxFracGenomeSubclone : float

        minSegmentBins : int

        altFracThreshold : float

        coverage : 

        maxCN : int

        txnE : float

        txnStrength : float

        normalizeMaleX : bool

        includeHOMD : bool

        fracReadsInChrYForMale : float

        chrXMedianForMale : float

        outDir : str

        libdir : str

        plotFileType : str

        plotYLim : tuple

        gender : dict

        genomeBuild : str

        genomeStyle : str

        chrs : array-like

        chrTrain : array-like

        chrNormalize : array-like

        verbose : bool

    Returns
    -------
        loglik : pandas.DataFrame
        
        results : dict
    """

    # Manage chr's
    if genomeStyle == "UCSC":
        chrTrain = np.array(["chr"+c for c in chrTrain])
        chrs = np.array(["chr"+c for c in chrs])

    normal_copy = None
    gender_mismatch = False
    numSamples = 1

    i = 0
    # Determine which chromosomes to use for training
    chrInd = tumour_copy.index.get_locs(chrTrain)

    ### RUN HMM ###
    start_time = time.perf_counter()

    results = {}
    loglik = pd.DataFrame([], columns=["init", "n_est", "phi_est", "BIC", "Frac_genome_subclonal",
                                    "Frac_CNA_subclonal", "loglik"],
                          index=np.arange(len(normal)* len(ploidy)))

    counter = 0
    compNames = np.repeat("NA", loglik.shape[0])
    mainName = np.repeat("NA", len(normal) * len(ploidy))

    #### restart for purity and ploidy values ####
    for n in normal:
        for p in ploidy:
            logR = tumour_copy.df.loc[:,"ratios"].copy(deep=True)

            param = hmm_utilities.getDefaultParameters(logR.loc[chrInd], maxCN=maxCN, includeHOMD=includeHOMD, 
                                        ct_sc=scStates, ploidy=p, e=txnE, e_sameState=50, strength=txnStrength)
            param["phi_0"] = np.repeat(p, numSamples)
            param["n_0"] = np.repeat(n, numSamples)

            ############################################
            ######## CUSTOM PARAMETER SETTINGS #########
            ############################################
            # 0.1x cfDNA #
            if lambda_p is None:
                logR_var = 1 / ((np.nanstd(logR, ddof=1, axis=0) / np.sqrt(len(param["ct"]))) **2)
                param["lambda"] = np.repeat(logR_var, len(param["ct"]))
                param["lambda"][np.in1d(param["ct"], np.array([2]))] = logR_var 
                param["lambda"][np.in1d(param["ct"], np.array([1,3]))] = logR_var 
                param["lambda"][param["ct"] >= 4] = logR_var / 5
                param["lambda"][param["ct"] == np.max(param["ct"])] = logR_var / 15
                param["lambda"][param["ct_sc_status"]] = logR_var / 10
            else:
                param["lambda"][np.in1d(param["ct"], np.array([2]))] = lambda_p[2]
                param["lambda"][np.in1d(param["ct"], np.array([1]))] = lambda_p[1]
                param["lambda"][np.in1d(param["ct"], np.array([3]))] = lambda_p[3]
                param["lambda"][param["ct"] >= 4] = lambda_p[4]
                param["lambda"][param["ct"] == np.max(param["ct"])] = lambda_p[2] / 15
                param["lambda"][param["ct_sc_status"]] = lambda_p[2] / 10
            param["alphaLambda"] = np.repeat(lambdaScaleHyperParam, len(param["ct"]))

            #############################################
            ################ RUN HMM ####################
            #############################################

            hmmResults_cor = hmm_utilities.HMMsegment(tumour_copy.copy(), None, dataType="ratios", 
                                                        param=param, chrTrain=chrTrain, maxiter=50,
                                                        estimateNormal=estimateNormal, estimatePloidy=estimatePloidy,
                                                        estimateSubclone=estimateScPrevalence, verbose=verbose)

            ########### CN correction ###################
            iteration = hmmResults_cor["results"]["iter"]

            # correct integer copy number based on estimated purity and ploidy
            correctedResults = hmm_utilities.correctIntegerCN(cn = hmmResults_cor["cna"].copy(),
                                                segs = hmmResults_cor["results"]["segs"].copy(), 
                                                purity = 1 - hmmResults_cor["results"]["n"][0][iteration],
                                                ploidy = hmmResults_cor["results"]["phi"][0][iteration],
                                                cellPrev = 1 - hmmResults_cor["results"]["sp"][0][iteration], 
                                                maxCNtoCorrect_autosomes = maxCN,
                                                maxCNtoCorrect_X = maxCN,
                                                minPurityToCorrect = 0.03, 
                                                gender = gender["gender"] if gender is not None else None,
                                                chrs = chrs,
                                                correctHOMD = includeHOMD)
            
            hmmResults_cor["results"]["segs"] = correctedResults["segs"]
            hmmResults_cor["cna"] = correctedResults["cn"]

            ## convert full diploid solution (of chrs to train) to have 1.0 normal or 0.0 purity
            ## check if there is an altered segment that has at least a minimum # of bins
            segsS = hmmResults_cor["results"]["segs"]
            segsS = segsS.loc[chrTrain, :]
            segAltInd = np.where(segsS.loc[:,"event"].values != "NEUT")[0]

            maxBinLength = -np.inf
            if np.sum(segAltInd) > 0:
                maxInd = np.argmax(segsS.ends()[segAltInd] - segsS.starts()[segAltInd] + 1)
                query = segsS.index[segAltInd[maxInd]]
                #query = query[query.unique_labels[0]]
                #subject = tumour_copy[pid]
                nhits = tumour_copy.nhits(query.start, query.end, query.label)
                maxBinLength = nhits
            
            ## check if there are proportion of total bins altered 
            # if segment size smaller than minSegmentBins, but altFrac > altFracThreshold, then still estimate TF
            cnaS = hmmResults_cor["cna"]
            altInd = cnaS.df.loc[np.in1d(cnaS.index.labels, chrTrain), "event"].values == "NEUT"
            altFrac = np.nansum(~altInd) / len(altInd)
            if maxBinLength <= minSegmentBins and altFrac <= altFracThreshold:
                hmmResults_cor["results"]["n"][0][iteration] = 1.0

            iteration = hmmResults_cor["results"]["iter"]
            results[counter] = hmmResults_cor
            loglik.loc[counter, "loglik"] = hmmResults_cor["results"]["loglik"][iteration]
            subClonalBinCount = pd.Series({0:hmmResults_cor["cna"].df.loc[:,"subclone_status"].sum()})
            fracGenomeSub = subClonalBinCount / pd.Series({0:hmmResults_cor["cna"].shape[0]})
            fracAltSub = subClonalBinCount / pd.Series({0:np.sum(hmmResults_cor["cna"].df.loc[:,"copy_number"].values != 2)})
            fracAltSub = pd.Series({i:0 if np.isnan(fracAltSub.loc[i]) else fracAltSub.loc[i] for i in fracAltSub.index.values})
            loglik.loc[counter, "Frac_genome_subclonal"] = ",".join(list(fracGenomeSub.values.astype(str)))
            loglik.loc[counter, "Frac_CNA_subclonal"] = ",".join(list(fracAltSub.values.astype(str)))
            loglik.loc[counter, "init"] = "n"+str(n)+"-p"+str(p)
            loglik.loc[counter, "n_est"] = ",".join(list(hmmResults_cor["results"]["n"][:, iteration].astype(str)))
            loglik.loc[counter, "phi_est"] = ",".join(list(hmmResults_cor["results"]["phi"][:, iteration].astype(str)))

            counter = counter + 1

    if verbose: print("Total ULP-WGS HMM Runtime:", ((time.perf_counter() - start_time) / 60), "min.")

    return loglik, results


"""Init for hmm."""

from __future__ import absolute_import
from .hmm import forward_backward, viterbi
from .hmm import get_include

# This is extracted automatically by the top-level setup.py.
__version__ = '0.1.0'


import numpy as np
import pandas as pd
from . import em_utilities
from .hmm import viterbi
from intervalframe import IntervalFrame
from ailist import LabeledIntervalArray


def getTransitionMatrix(K, e, strength):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    A = np.zeros((K,K), order="F")

    for j in range(K):
        A[j,:] = (1 - e) / (K - 1)
        A[j,j] = e

    A = normalize(A)
    A_prior = A
    dirPrior = A * strength

    return {"A":A, "dirPrior":dirPrior}


def normalize(A):
    """
    Normalize a given array to sum to 1

    Parameters
    ----------

    
    Returns
    -------


    """

    def vectorNormalize(x):
        return x / (np.sum(x, axis=0) + (np.sum(x, axis=0) == 0))

    if A.ndim < 2:
        M = vectorNormalize(A)
    else:
        M = vectorNormalize(A)

    return M


def getDefaultParameters(logR, maxCN = 5, ct_sc = None, ploidy = 2, e = 0.9999999,
                         e_sameState = 10, strength = 10000000, includeHOMD = False):
    """
    Determine default parameters
    
    Parameters
    ----------
        logR : pandas.Series
            Log ratios
        maxCN : int
            Maximum Copy Numbers
        ct_sc : int
            
        ploidy : int
            Assumed ploidy
        e : float
            
        e_sameState : int
            
        strength : int
            (default:10000000)
        includeHOMD : bool
            Whether to include homozygous deletions
        
    Returns
    -------
        param : dict
            Dictionary of HMM starting parameters
    
    """
    
    # Determine if HOMD is included
    if includeHOMD:
        ct = np.arange(0, maxCN+1)
    else:
        ct = np.arange(1, maxCN+1)
    
    # Initialize parameters
    length_ct_sc = 0 if ct_sc is None else len(ct_sc)
    param = {"strength" : strength, "e" : e,
    		"ct" : ct if ct_sc is None else np.append(ct, ct_sc),
    		"ct_sc_status" : np.append(np.repeat(False, len(ct)), np.repeat(True, length_ct_sc)),
    		"phi_0" : 2, "alphaPhi" : 4, "betaPhi" : 1.5,
    		"n_0" : 0.5, "alphaN" : 2, "betaN" : 2,
    		"sp_0" : 0.5, "alphaSp" : 2, "betaSp" : 2,
    		"lambda" : np.repeat(100, len(ct) + length_ct_sc).reshape(-1,1),
    		"nu" : 2.1,
    		"kappa" : np.repeat(75, len(ct)), 
    		"alphaLambda" : 5}
            
    K = len(param["ct"])
    
    # Initialize hyperparameters for precision using observed data
    param["numberSamples"] = 1
    betaLambdaVal = ((np.nanstd(logR, ddof=1) / np.sqrt(len(param["ct"]))) **2)
    param["betaLambda"] = np.repeat(betaLambdaVal, len(param["ct"]))

    param["alphaLambda"] = np.repeat(param["alphaLambda"], K)

    S = param["numberSamples"]

    logR_var = 1 / ((np.nanstd(logR, ddof=1, axis=0) / np.sqrt(len(param["ct"]))) **2)

    if logR.ndim == 2 and logR.shape[1] > 1:
        param["lambda"] = np.tile(logR_var, (K, S))
    else:
        param["lambda"] = np.repeat(logR_var, len(param["ct"]))
        param["lambda"][np.in1d(param["ct"], np.array([2]))] = logR_var 
        param["lambda"][np.in1d(param["ct"], np.array([1,3]))] = logR_var 
        param["lambda"][param["ct"] >= 4] = logR_var / 5
        param["lambda"][param["ct"] == np.max(param["ct"])] = logR_var / 15
        param["lambda"][param["ct_sc_status"]] = logR_var / 10

    # Define joint copy number states 
    param["jointCNstates"] = pd.DataFrame(np.repeat(param["ct"], S).astype(int), columns=["Sample_1"])
    param["jointSCstatus"] = pd.DataFrame(np.repeat(param["ct_sc_status"], S), columns=["Sample_1"])

    # Initialize transition matrix to the prior
    txn = getTransitionMatrix(K**S, e, strength)
    cnStateDiff = param["jointCNstates"].apply(lambda x: (np.abs(np.max(x) - np.min(x))), axis=1).values

    if e_sameState > 0 and S > 1:
        txn["A"][:, cnStateDiff==0] = txn["A"][:, cnStateDiff==0] * e_sameState * K 
        txn["A"][:, cnStateDiff>=3] = txn["A"][:, cnStateDiff>=3] / e_sameState / K

    for i in range(txn["A"].shape[0]):
        for j in range(txn["A"].shape[1]):
            if i == j:
                txn["A"][i, j] = e

    txn["A"] = normalize(txn["A"])
    param["A"] = txn["A"]
    param["dirPrior"] = txn["A"] * strength

    param["A"][:, param["ct_sc_status"]] = param["A"][:, param["ct_sc_status"]] / 10
    param["A"] = normalize(param["A"])
    param["dirPrior"][:, param["ct_sc_status"]] = param["dirPrior"][:, param["ct_sc_status"]] / 10

    if includeHOMD:
        K = len(param["ct"])
        param["A"][0, 1:K] = param["A"][0, 1:K] * 1e-5
        param["A"][1:K, 0] = param["A"][1:K, 0] * 1e-5
        param["A"][0, 0] = param["A"][0, 0] * 1e-5
        param["A"] = normalize(param["A"])
        param["dirPrior"] = param["A"] * param["strength"]

    param["kappa"] = np.repeat(75, K**S)
    param["kappa"][cnStateDiff == 0] = param["kappa"][cnStateDiff == 0] + 125
    param["kappa"][cnStateDiff >=3] = param["kappa"][cnStateDiff >=3] - 50
    param["kappa"][np.sum(param["jointCNstates"].values==2, axis=1) == S] = 800

    return param


def runViterbi(convergedParams, chroms, verbose=False):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    if verbose: print("runViterbi: Segmenting and classifying")
    
    indexes = np.unique(chroms, return_index=True)[1]
    chrs = np.array([chroms[index] for index in sorted(indexes)])
    chrsI = []

    # initialise the chromosome index and the init state distributions
    for i in range(len(chrs)):
        chrsI.append(np.where(chroms == chrs[i])[0])

    segs = []
    py = convergedParams["py"]
    N = py.shape[1]
    Z = np.zeros(N, dtype=int)
    convergeIter = convergedParams["iter"]
    piG = convergedParams["pi"][:, convergeIter]
    A = convergedParams["A"]


    for c in range(len(chrsI)):
        I = chrsI[c]
        output = viterbi(np.log(piG), np.log(A), np.log(py[:, I]))
        Z[I] = output["path"]
        segs.append(output["seg"])

    return {"segs":segs, "states":Z}
        

def merge_Trues(x):
    """


    Parameters
    ----------
        x : numpy.ndarray
    
    Returns
    -------
        Generator of tuples

    """
    
    # Merge nearby True values
    x1 = np.hstack([ [False], x, [False] ])  # padding
    d = np.diff(x1.astype(int))
    starts = np.where(d == 1)[0]
    ends = np.where(d == -1)[0]

    # Determine indices
    for segment in zip(starts,ends):
        yield segment


def segment_data(x, states, convergedParams):
    """


    Parameters
    ----------
        x : IntervalFrame
    
    Returns
    -------
        total_segList : IntervalFrame

    """

    if np.sum(convergedParams["param"]["ct"] == 0) == 0:
        includeHOMD = False
    else:
        includeHOMD = True
    
    if ~includeHOMD:
        names = ["HETD","NEUT","GAIN","AMP","HLAMP"] + ["HLAMP"+str(i) for i in range(2,25)]
    else:
        names = ["HOMD","HETD","NEUT","GAIN","AMP","HLAMP"] + ["HLAMP"+str(i) for i in range(2,25)]
    
    jointStates = convergedParams["param"]["jointCNstates"].values

    #indexes = np.sort(np.unique(x.labels(), return_index=True)[1])
    chroms = x.index.unique_labels
    sample_segList = []
    sample_intervals = {}

    sample_intervals = LabeledIntervalArray()
    for i in range(len(chroms)):
        chrom = chroms[i]
        #sample_intervals[chrom] = LabeledIntervalArray()
        chrom_selected = x.loc[chrom,:]
        chrom_positions = chrom_selected.starts
        chrom_values = chrom_selected.df["ratios"].values
        chrom_states = states[x.index.get_locs([chrom])]
        chrom_segList = []
        for state in np.unique(chrom_states):
            name = names[int(state)]
            for start, end in merge_Trues(chrom_states==state):
                end = min(end, len(chrom_positions)-1)
                median = np.nanmedian(chrom_values[start:end])
                #chrom_segList.append([chrom, chrom_positions[start], chrom_positions[end], int(state), name, median, jointStates[int(state),j]])
                sample_intervals.add(chrom_positions[start], chrom_positions[end], str(chrom))
                chrom_segList.append([int(state), name, median, jointStates[int(state)]])

        #chrom_segList.sort(key = lambda x: x[1])
        sample_segList += chrom_segList

    #total_segList.append(pd.DataFrame(sample_segList, columns=["chrom", "start", "end", "state", "event", "median", "copy_number"]))
    total_segList = IntervalFrame(sample_intervals,
                                  pd.DataFrame(sample_segList,
                                               columns=["state", "event", "median", "copy_number"]))

    return total_segList
    

def HMMsegment(x, validInd=None, dataType="ratios", param=None, 
               chrTrain=np.arange(22).astype(str), maxiter=50, estimateNormal=True, estimatePloidy=True, 
               estimatePrecision=True, estimateSubclone=True, estimateTransition=True,
               estimateInitDist=True, logTransform=False, verbose=False):
    """


    Parameters
    ----------
        x : IntervalFrame

        validInd : array-like

        dataType : str

        param : dict

        chrTrain : array-like

        maxiter : int

        estimateNormal : bool

        estimate Ploidy : bool

        estimatePrecision : bool

        estimateSubclone : bool

        estimateTransition : bool

        estimateInitDist : bool

        logTransform : bool

        verbose : bool
    
    Returns
    -------
        dict

    """
    
    # Find chromosomes present
    chroms = x.index.labels

    # Setup columns
    dataMat = x.df.loc[:,dataType].to_frame()

    # Normalize by median and log data
    if logTransform:
        dataMat = dataMat.apply(lambda x: np.log(x / np.nanmedian(x)), axis=1)
    else:
        dataMat = np.log(2**dataMat)

    # Update variable x with loge instead of log2
    x.df.loc[:, dataType] = dataMat.iloc[:, 0]

    chrInd = np.in1d(chroms, chrTrain)
    if validInd is not None:
        chrInd = np.logical_and(chrInd, validInd)
    if param is None:
        param = getDefaultParameters(dataMat.loc[chrInd,:])

    ####### RUN EM ##########

    convergedParams = em_utilities.runEM(dataMat, chroms, chrInd, param, maxiter, 
                                        verbose, estimateNormal=estimateNormal, estimatePloidy=estimatePloidy, 
                                        estimateSubclone=estimateSubclone, estimatePrecision=estimatePrecision, 
                                        estimateTransition=estimateTransition, estimateInitDist=estimateInitDist)
    
    viterbiResults = runViterbi(convergedParams, chroms, verbose=verbose)

    # Segment data based on called states
    segs = segment_data(x, viterbiResults["states"], convergedParams)
    
    names = np.array(["HOMD","HETD","NEUT","GAIN","AMP","HLAMP"] + ["HLAMP"+str(i) for i in range(2,25)])

    copyNumber = param["jointCNstates"].values[viterbiResults["states"]]
    subclone_status = param["jointSCstatus"].values[viterbiResults["states"]]

    chroms = x.index.labels

    cnaList = IntervalFrame.from_array(x.starts, x.ends, labels=chroms)
    cnaList.loc[:,"copy_number"] = copyNumber
    cnaList.loc[:,"event"] = names[copyNumber]
    cnaList.loc[:,"logR"] = np.log2(np.exp(dataMat.values))
    cnaList.loc[:,"subclone_status"] = subclone_status

    ## segment mean loge -> log2
    segs.df["median"] = np.log2(np.exp(segs.df["median"].values))
    ## add subclone status
    segs.df["subclone_status"] = param["jointSCstatus"].values[segs.df["state"].values]
    
    convergedParams["segs"] = segs

    return {"cna":cnaList, "results":convergedParams, "viterbiResults":viterbiResults}


def logRbasedCN(x, purity, ploidyT, cellPrev=np.nan, cn = 2):
    """
    compute copy number using corrected log ratio

    Parameters
    ----------

    
    Returns
    -------


    """

    if len(cellPrev) == 1 and np.isnan(cellPrev):
        cellPrev = 1
    else: #if cellPrev is a vector
        cellPrev[np.isnan(cellPrev)] = 1
    
    ct = (2**x 
            * (cn * (1 - purity) + purity * ploidyT * (cn / 2)) 
            - (cn * (1 - purity)) 
            - (cn * purity * (1 - cellPrev)))
    ct = ct / (purity * cellPrev)
    ct = np.maximum(ct, 1/2**6)

    # Replace inf with 0
    ct[ct == np.inf] = 0
    
    return ct



def correctIntegerCN(cn, segs, purity, ploidy, cellPrev, callColName = "event", maxCNtoCorrect_autosomes = None, 
		             maxCNtoCorrect_X = None, correctHOMD = True, minPurityToCorrect = 0.2, gender = "male",
                     chrs = np.append(np.arange(1,23).astype(str), np.array(["X"]))):
    """
    Recompute integer CN for high-level amplifications
    compute logR-corrected copy number

    Parameters
    ----------

    
    Returns
    -------


    """

    names = np.array(["HOMD","HETD","NEUT","GAIN","AMP","HLAMP"] + ["HLAMP"+str(i) for i in range(2,3000)] + ["NA"])

    ## set up chromosome style
    autosomeStr = np.array([c for c in chrs if "X" not in c and "Y" not in c])
    chrXStr = [c for c in chrs if "X" in c][0]

    if maxCNtoCorrect_autosomes is None:
        maxCNtoCorrect_autosomes = np.nanmax(segs.loc[np.in1d(segs.loc[:,"chrom"].values, autosomeStr), "copy_number"].values)
    
    if maxCNtoCorrect_X is None and gender == "female" and len(chrXStr) > 0:
        maxCNtoCorrect_X = np.nanmax(segs.loc[segs.loc[:,"chrom"].values == chrXStr, "copy_number"])
    
    ## correct log ratio and compute corrected CN
    cellPrev_seg = np.ones(segs.shape[0])
    cellPrev_seg[segs.df["subclone_status"]] = cellPrev
    segs.loc[:,"logR_Copy_Number"] = logRbasedCN(segs.loc[:,"median"].values, purity, ploidy, cellPrev_seg, cn=2)
    if gender == "male" and len(chrXStr) > 0: ## analyze chrX separately
        ind_cnChrX = np.where(segs.loc[:,"chrom"].values == chrXStr)[0]
        segs.loc[ind_cnChrX,"logR_Copy_Number"] = logRbasedCN(segs.loc[:,"median"].values[ind_cnChrX], purity, ploidy, cellPrev_seg[ind_cnChrX], cn=1)
    
    ## assign copy number to use - Corrected_Copy_Number
    # same ichorCNA calls for autosomes - no change in copy number
    segs.loc[:,"Corrected_Copy_Number"] = segs.loc[:,"copy_number"].values.astype(int)
    segs.loc[:,"Corrected_Call"] = segs.loc[:,callColName].values

    ind_change = np.array([])
    if purity >= minPurityToCorrect:
        # ichorCNA calls adjusted for >= copies - HLAMP
        # perform on all chromosomes
        ind_cn = np.where(np.logical_or(segs.loc[:,"copy_number"].values >= maxCNtoCorrect_autosomes,
                                        segs.loc[:,"logR_Copy_Number"].values >= maxCNtoCorrect_autosomes * 1.2))[0]
        segs.df.loc[ind_cn,"Corrected_Copy_Number"] = np.round(segs.df.loc[:,"logR_Copy_Number"].values[ind_cn]).astype(int)
        segs.df.loc[ind_cn,"Corrected_Call"] = names[segs.df.loc[:,"Corrected_Copy_Number"].values[ind_cn] + 1]
        ind_change = np.append(ind_change, ind_cn)

        # ichorCNA calls adjust for HOMD
        if correctHOMD:
            ind_cn = np.where(np.logical_and(np.in1d(segs.index.labels, chrs),
                                             np.logical_or(segs.loc[:,"copy_number"].values == 0,
                                                           segs.loc[:,"logR_Copy_Number"].values == 1/2**6)))[0]
            segs.df.loc[ind_cn,"Corrected_Copy_Number"] = np.round(segs.df.loc[:,"logR_Copy_Number"].values[ind_cn])
            segs.df.loc[ind_cn,"Corrected_Call"] = names[segs.df.loc[:,"Corrected_Copy_Number"].values[ind_cn] + 1]
            ind_change = np.append(ind_change, ind_cn)
        
        # Re-adjust chrX copy number for males (females already handled above)
        if gender == "male" and len(chrXStr) > 0:
            ind_cn = np.where(np.logical_and(segs.index.labels == chrXStr,
                                            np.logical_or(segs.df.loc[:,"copy_number"].values >= maxCNtoCorrect_X,
                                                          segs.df.loc[:,"logR_Copy_Number"].values >= maxCNtoCorrect_X * 1.2)))[0]
            segs.df.loc[ind_cn,"Corrected_Copy_Number"] = np.round(segs.df.loc[:,"logR_Copy_Number"].values[ind_cn]).astype(int)
            segs.df.loc[ind_cn,"Corrected_Call"] = names[segs.df.loc[:,"Corrected_Copy_Number"].values[ind_cn] + 2]
            ind_change = np.append(ind_change, ind_cn)
        
    ## adjust the bin level data ##
    # 1) assign the original calls
    cn.loc[:,"Corrected_Copy_Number"] = cn.loc[:,"copy_number"].values
    cn.loc[:,"Corrected_Call"] = cn.loc[:,callColName].values
    cellPrev_cn =  np.ones(cn.shape[0])
    cellPrev_cn[cn.loc[:,"subclone_status"].values] = cellPrev
    cn.loc[:,"logR_Copy_Number"] = logRbasedCN(cn.loc[:,"logR"].values, purity, ploidy, cellPrev_cn, cn=2)
    if gender == "male" and len(chrXStr) > 0: ## analyze chrX separately
        ind_cnChrX = np.where(cn.loc[:,"chr"].values == chrXStr)[0]
        cn.loc[ind_cnChrX,"logR_Copy_Number"] = logRbasedCN(cn.loc[:,"logR"].values[ind_cnChrX], purity, ploidy, cellPrev_cn[ind_cnChrX], cn=1)
    
    # 2) find which segs changed/adjusted
    ind_change = np.unique(ind_change).astype(int)

    # 3) correct bins overlapping adjusted segs
    hits = np.array([])
    for i in ind_change:
        fields = segs.iloc[i,:]

        overlaps = cn.index.intersect(fields.index[0].start,
                                      fields.index[0].end,
                                      fields.index[0].label,
                                      return_intervals=False,
                                      return_index=True)

        cn.df.loc[:,"Corrected_Copy_Number"].values[overlaps] = fields.df.loc[:,"Corrected_Copy_Number"].values
        cn.df.loc[:,"Corrected_Call"].values[overlaps] = fields.df.loc[:,"Corrected_Call"].values
        hits = np.append(hits, overlaps)

    # 4) correct bins that are missed as high level amplifications
    ind_cn = np.where(np.logical_or(cn.df.loc[:,"copy_number"].values >= maxCNtoCorrect_autosomes,
                                    cn.df.loc[:,"logR_Copy_Number"].values >= maxCNtoCorrect_autosomes * 1.2))[0]
    ind_cn = ind_cn[~np.in1d(ind_cn, hits)]
    cn.df.loc[ind_cn,"Corrected_Copy_Number"] = np.round(cn.df.loc[:,"logR_Copy_Number"].values[ind_cn])
    corr_copy = np.round(cn.df.loc[:,"logR_Copy_Number"].values[ind_cn])
    corr_copy[pd.isnull(corr_copy)] = -2
    corr_copy[corr_copy >= len(names)] = len(names) - 2 # scale to max label
    corr_copy = corr_copy.astype(int)
    cn.df.loc[ind_cn,"Corrected_Call"] = names[corr_copy + 1]
    
    return {"cn":cn, "segs":segs}


import cython

import numpy as np
cimport numpy as np
import os


# Declare interface to C code
cdef extern from "utilities.c":
    # forward_backward utilities
    void transposeSquareInPlace(double *out_m, double *in_m, int K)
    void multiplyInPlace(double *result, double *u, double *v, int K)
    double normalizeInPlace(double *A, int N)
    void multiplyMatrixInPlace(double *result, double *trans, double *v, int K)
    void outerProductUVInPlace(double *out, double *u, double *v, int K)
    void componentVectorMultiplyInPlace(double *out, double *u, double *v, int L)

    # viterbi utilities
    void addVectors(double *out, double *u, double *v, int L)
    void setVectorToValue_int(double *A, int value, int L)
    void maxVectorInPlace(double *out_value, int *out_index, double *A, int L)


def get_include():
    """
    Get file directory if C headers
    
    Arguments:
    ---------
        None
    Returns:
    ---------
        str (Directory to header files)
    """

    return os.path.split(os.path.realpath(__file__))[0]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def forward_backward(np.ndarray[double, ndim=1, mode="fortran"] piZ,
                     np.ndarray[double, ndim=2, mode="fortran"] A,
                     np.ndarray[double, ndim=2, mode="fortran"] py):
    """
    """

    cdef double[:] init_state_distrib = piZ
    cdef double[:,:] transmat = A
    cdef double[:,:] obslik = py

    cdef int K = piZ.size
    cdef double[:] transmatT_data = np.zeros(K*K, dtype=np.double, order="F")
    transposeSquareInPlace(&transmatT_data[0], &transmat[0,0], K)

    cdef int T = py.shape[1]

    cdef double[:] scale_data = np.zeros(T, dtype=np.double, order="F")
    cdef double[:,:] alpha_data = np.zeros((K, T), dtype=np.double, order="F")
    cdef double[:,:] beta_data = np.zeros((K, T), dtype=np.double, order="F")
    cdef double[:,:] gamma_data = np.zeros((K, T), dtype=np.double, order="F")
    cdef double[:] loglik_data = np.zeros(1, dtype=np.double, order="F")

    # ********* FOWARD **********
    cdef int t = 0
    multiplyInPlace(&alpha_data[0,t], &init_state_distrib[0], &obslik[0,t], K)
    scale_data[t] = normalizeInPlace(&alpha_data[0,t], K)

    cdef double[:] m_data = np.zeros(K, dtype=np.double, order="F")

    for t in range(1,T):
        multiplyMatrixInPlace(&m_data[0], &transmatT_data[0], &alpha_data[0,t-1], K)
        multiplyInPlace(&alpha_data[0,t], &m_data[0], &obslik[0,t], K)
        scale_data[t] = normalizeInPlace(&alpha_data[0,t], K)

    loglik_data[0] = 0
    for t in range(T):
        loglik_data[0] += np.log(scale_data[t])

    # ********* BACKWARD **********

    t = T - 1
    for d in range(K):
        beta_data[:,t] = 1
        gamma_data[:,t] = alpha_data[:,t]

    cdef double[:] b_data = np.zeros(K, dtype=np.double, order="F")
    cdef double[:,:,:] eta_data = np.zeros((K,K,T), dtype=np.double, order="F")
    cdef double[:,:] squareSpace_data = np.zeros((K,K), dtype=np.double, order="F")
    
    # We have to remember that the 1:T range in R is 0:(T-1) in C
    for t in range(T-2, -1, -1):

        # setting beta
        multiplyInPlace(&b_data[0], &beta_data[0,t+1], &obslik[0,t+1], K)
        multiplyMatrixInPlace(&m_data[0], &A[0,0], &b_data[0], K)
        normalizeInPlace(&m_data[0], K)

        for d in range(K):
            beta_data[d,t] = m_data[d]
        # using "m" again as valueholder

        # setting eta, whether we want it or not in the output
        outerProductUVInPlace(&squareSpace_data[0,0], &alpha_data[0,t], &b_data[0], K)
        componentVectorMultiplyInPlace(&eta_data[0,0,t], &transmat[0,0], &squareSpace_data[0,0], K * K)
        normalizeInPlace(&eta_data[0,0,t], K * K)

        # Setting gamma
        multiplyInPlace(&m_data[0], &alpha_data[0,t], &beta_data[0,t], K)
        normalizeInPlace(&m_data[0], K)
        for d in range(K):
            gamma_data[d,t] = m_data[d]

    results = {}
    results["rho"] = np.asarray(gamma_data)
    results["alpha"] = np.asarray(alpha_data)
    results["beta"] = np.asarray(beta_data)
    results["xi"] = np.asarray(eta_data)
    results["loglik"] = np.asarray(loglik_data)
    
    return results


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def viterbi(np.ndarray[double, ndim=1, mode="fortran"] piZ,
            np.ndarray[double, ndim=2, mode="fortran"] A,
            np.ndarray[double, ndim=2, mode="fortran"] py):
    """
    """

    cdef int K = piZ.size

    if A.shape[0] != K or A.shape[1] != K:
        raise IndexError("The transition matrix must be of size %d x %d", K, K)
    
    cdef int T = py.shape[1]

    if py.shape[0] != K:
        raise IndexError("The observed likelihoods must have %d rows", K)
        
    cdef double[:,:] delta_data = np.zeros((K, T), dtype=np.double, order="F")
    cdef double[:] d_data = np.zeros(K, dtype=np.double, order="F")
    cdef double[:] loglik_data = np.zeros(1, dtype=np.double, order="F")
    cdef int[:,:] psi_data = np.zeros((K, T), dtype=np.intc, order="F")
    cdef int[:] path_data = np.zeros(T, dtype=np.intc, order="F")

    cdef int t = 0
    #addVectors(delta + t * K, prior, obslik + t * K, K);
    addVectors(&delta_data[0,t], &piZ[0], &py[0,t], K)
    #setVectorToValue_int(psi + t * K, 0, K);
    #setVectorToValue_int(&psi_data[0,t], 0, K)

    # FORWARD
    for t in range(1,T):
        for j in range(K):
            #addVectors(d, delta + (t - 1) * K, transmat + j * K, K);
            addVectors(&d_data[0], &delta_data[0,t-1], &A[0,j], K)
            #maxVectorInPlace(delta + j + t * K, psi + j + t * K, d, K);
            maxVectorInPlace(&delta_data[j,t], &psi_data[j,t], &d_data[0], K)
            #delta[j + t * K] += obslik[j + t * K];
            delta_data[j,t] += py[j,t]
    
    # BACKWARD
    t = T - 1
    #maxVectorInPlace(d, path + t, delta + t * K, K); // Use d[0] as temp variable
    maxVectorInPlace(&d_data[0], &path_data[t], &delta_data[0,t], K) # Use d[0] as temp variable
    loglik_data[0] = d_data[0]

    for t in range(T-2, -1, -1):
        #path[t] = psi[path[t+1] + (t+1)*K]
        path_data[t] = psi_data[path_data[t+1],t+1]
    
    cdef double[:,:] changes_data = np.zeros((4, T), dtype=np.double, order="F")
    changesCounter = 0
    cdef int ind1
    cdef int ind2
    ind1, ind2 = np.unravel_index(changesCounter + 0 * T, (4,T), order="F")
    changes_data[ind1, ind2] = 0
    ind = np.unravel_index(changesCounter + 1 * T, (4,T), order="F")
    changes_data[ind[0], ind[1]] = 0 # overwritten
    ind = np.unravel_index(changesCounter + 2 * T, (4,T), order="F")
    changes_data[ind[0], ind[1]] = path_data[0]
    ind = np.unravel_index(changesCounter + 3 * T, (4,T), order="F")
    changes_data[ind[0], ind[1]] = 0
    changesCounter = 1

    for t in range(1,T):
        if path_data[t] != path_data[t-1]:
            ind = np.unravel_index(changesCounter + 0 * T, (4,T), order="F")
            changes_data[ind[0], ind[1]] = t
            ind = np.unravel_index((changesCounter - 1) + 1 * T, (4,T), order="F")
            changes_data[ind[0], ind[1]] = t - 1
            ind = np.unravel_index(changesCounter + 2 * T, (4,T), order="F")
            changes_data[ind[0], ind[1]] = path_data[t]
            ind = np.unravel_index(changesCounter + 3 * T, (4,T), order="F")
            changes_data[ind[0], ind[1]] = 0
            changesCounter+=1

    ind = np.unravel_index((changesCounter - 1) + 1 * T, (4,T), order="F")
    changes_data[ind[0], ind[1]] = T - 1

    # Reformat to segs
    cdef double[:,:] segs_data = np.zeros((changesCounter, 4), dtype=np.double, order="F")
    for t in range(changesCounter):
        seg_ind = np.unravel_index(t + 0 * changesCounter, (changesCounter, 4), order="F")
        ind = np.unravel_index(t + 0 * T, (4,T), order="F")
        segs_data[seg_ind[0], seg_ind[1]] = changes_data[ind[0], ind[1]] + 1.0

        seg_ind = np.unravel_index(t + 1 * changesCounter, (changesCounter, 4), order="F")
        ind = np.unravel_index(t + 1 * T, (4,T), order="F")
        segs_data[seg_ind[0], seg_ind[1]] = changes_data[ind[0], ind[1]] + 1.0
        
        seg_ind = np.unravel_index(t + 2 * changesCounter, (changesCounter, 4), order="F")
        ind = np.unravel_index(t + 2 * T, (4,T), order="F")
        segs_data[seg_ind[0], seg_ind[1]] = changes_data[ind[0], ind[1]] + 1.0
        
        seg_ind = np.unravel_index(t + 3 * changesCounter, (changesCounter, 4), order="F")
        ind = np.unravel_index(t + 3 * T, (4,T), order="F")
        segs_data[seg_ind[0], seg_ind[1]] = changes_data[ind[0], ind[1]]
    
    results = {}
    results["path"] = np.asarray(path_data)
    results["loglik"] = np.asarray(loglik_data)
    results["seg"] = np.asarray(segs_data)

    return results


import numpy as np
import pandas as pd
from math import lgamma
from scipy.special import betaln, gammaln
from . import hmm_utilities
from scipy.special import gamma
from .hmm import forward_backward
from scipy.optimize import minimize, Bounds
import time


def get2and3ComponentMixture(ct, ct_sc_status, n, sp, phi):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    #S = len(n)
    cn = 2
    mu = np.tile(np.nan, ct.shape)
    #for s in range(S):
        #subclonal 3 component
    mu[ct_sc_status == True] = (((1 - n) * (1 - sp) * ct[ct_sc_status == True]) + 
                ((1 - n) * sp * cn) + (n * cn)) / ((1 - n) * phi + n * cn)
    #clonal 2 component
    mu[ct_sc_status == False] = ((1 - n) * ct[ct_sc_status == False] + n * cn) / ((1 - n) * phi + n * cn)
    
    return np.log(mu)


def tdistPDF(x, mu, lambda_p, nu):
    """


    Parameters
    ----------
        x : np.ndarrau
    
    Returns
    -------


    """

    S = x.shape[1]
    if S is not None:
        p = None
        for s in range(S):
            tpdf = (gamma(nu / 2 + 0.5) / gamma(nu / 2)) * ((lambda_p[s] / (np.pi * nu)) **(0.5)) * (1 + (lambda_p[s] * (x[:, s] - mu[s]) **2) / nu) **(-0.5 * nu - 0.5)
            p = tpdf if p is None else np.array([p, tpdf])
    else:
        nu = np.repeat(nu, len(mu))
        p = (gamma(nu / 2 + 0.5) / gamma(nu / 2)) * ((lambda_p / (np.pi * nu)) **(0.5)) * (1 + (lambda_p * (x - mu) **2) / nu) **(-0.5 * nu - 0.5)
    
    p[np.isnan(p)] = 1

    #nu = np.repeat(nu, len(mu))
    #p = (gamma(nu / 2 + 0.5) / gamma(nu / 2)) * ((lambda_p / (np.pi * nu)) **(0.5)) * (1 + (lambda_p * (x - mu) **2) / nu) **(-0.5 * nu - 0.5)
    #p[np.isnan(p)] = 1
    
    return p


def dirichletpdf(x, alpha, verbose=False):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    if (x < 0).any():
        return 0
    
    if np.abs(np.sum(x) - 1) > 1e-3:
        if verobose: print("Dirichlet PDF: probabilities do not sum to 1")
        return 0
    
    p = np.exp(gammaln(np.sum(alpha)) - np.sum(gammaln(alpha))) * np.prod(x **(alpha - 1))
    return p


def dirichletpdflog(x, k):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    c = gammaln(np.nansum(k)) - np.nansum(gammaln(k))  #normalizing constant
    l = np.nansum((k - 1) * np.log(x))  #likelihood
    y = c + l

    return y


def gammapdflog(x, a, b): #rate and scale parameterization
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    c = a * np.log(b) - gammaln(a)  # normalizing constant  
    l = (a - 1) * np.log(x) + (-b * x)  #likelihood  
    y = c + l

    return y


def betapdflog(x, a, b):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    y = -betaln(a, b) + (a - 1) * np.log(x) + (b - 1) * np.log(1 - x)

    return y


def estimateMixWeightsParamMap(rho, kappa):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    K = rho.shape[0]
    pi = (np.nansum(rho, axis=1) + kappa - 1) / (np.sum(np.nansum(rho, axis=1)) + np.sum(kappa) - K)

    return pi


def stLikelihood(n, sp, phi, lambda_p, params, D, rho):
    """
    Student's t likelihood function

    Parameters
    ----------

    
    Returns
    -------


    """

    KS = rho.shape[0]
    lik = 0
    
    # Recalculate the likelihood
    lambdaKS = lambda_p.reshape(-1,1)
    mus = get2and3ComponentMixture(params["jointCNstates"].values, params["jointSCstatus"].values, n, sp, phi)
    for ks in range(KS):
        probs = np.log(tdistPDF(D.values, mus[ks,:], lambdaKS[ks,:], params["nu"]))    
        # multiply across samples for each data point to get joint likelihood.
        l = np.dot(rho[ks,:], probs)
        lik = lik + l
    
    return lik


def completeLikelihoodFun(x, pType, n, sp, phi, lambda_p, piG, A, params, D, rho, Zcounts,
                                  estimateNormal = True, estimatePloidy = True,
                                  estimatePrecision = True, estimateTransition = False,
                                  estimateInitDist = True, estimateSubclone = True):
    """


    Parameters
    ----------

    
    Returns
    -------


    """
    
    KS = rho.shape[0]
    N = rho.shape[1]
    S = params["numberSamples"]
    K = len(params["ct"])

    lambda_p = np.tile(lambda_p, S)

    if estimatePrecision and np.sum(pType == "lambda") > 0:
        lambda_p = np.tile(x[pType == "lambda"], S)

    if estimateNormal and np.sum(pType == "n") > 0:
        n = x[pType == "n"]

    if estimateSubclone and np.sum(pType == "sp") > 0:
        sp = x[pType == "sp"]
    
    if estimatePloidy and np.sum(pType == "phi") > 0:
        phi = x[pType == "phi"]

    ## prior probabilities ##
    prior = priorProbs(n, sp, phi, lambda_p, piG, A, params, 
                        estimateNormal = estimateNormal, estimatePloidy = estimatePloidy,
                        estimatePrecision = estimatePrecision, estimateTransition = estimateTransition,
                        estimateInitDist = estimateInitDist, estimateSubclone = estimateSubclone)
    ## likelihood terms ##
    likObs = stLikelihood(n, sp, phi, lambda_p, params, D, rho)
    likInit = np.dot(rho[:, 1], np.log(piG))
    likTxn = 0
    for ks in range(KS):
        likTxn = likTxn + np.dot(Zcounts[ks,:], np.log(A[ks,:]))

    ## sum together ##
    lik = likObs + likInit + likTxn
    prior = prior["prior"]
    f = lik + prior
    
    # Multiply by -1, because we want to maximize, not minimize
    return f * -1


def priorProbs(n, sp, phi, lambda_p, piG, A, params, 
                estimateNormal = True, estimatePloidy = True,
                estimatePrecision = True, estimateTransition = True,
                estimateInitDist = True, estimateSubclone = True):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    S = params["numberSamples"]
    K = len(params["ct"])
    KS = K **S
    
    ## prior terms ##
    priorA = 0
    if estimateTransition:
        for ks in range(KS):
            priorA = priorA + dirichletpdflog(A[ks,:], params["dirPrior"][ks,:])

    priorLambda = 0
    if estimatePrecision:
        if S > 1:
            for s in range(S):
                for k in range(K):
                    priorLambda = priorLambda + gammapdflog(lambda_p[k, s], params["alphaLambda"][k], params["betaLambda"][k, s])
        else:
            for k in range(K):
                priorLambda = priorLambda + gammapdflog(lambda_p[k], params["alphaLambda"][k], params["betaLambda"][k])

    priorLambda = priorLambda
    priorN = 0
    if estimateNormal:
        priorN = np.sum(betapdflog(n, params["alphaN"], params["betaN"]))

    priorSP = 0
    if estimateNormal:
        priorSP = np.sum(betapdflog(sp, params["alphaSp"], params["betaSp"]))
    
    priorPhi = 0
    if estimatePloidy:
        priorPhi = np.sum(gammapdflog(phi, params["alphaPhi"], params["betaPhi"]))
    
    priorPi = 0
    if estimateInitDist:
        priorPi = dirichletpdflog(piG, params["kappa"])
    
    prior = priorA + priorLambda + priorN + priorSP + priorPhi + priorPi

    return {"prior":prior, "priorA":priorA, "priorLambda":priorLambda, 
            "priorN":priorN, "priorSP":priorSP, "priorPhi":priorPhi, "priorPi":priorPi}


def estimateParamsMap(D, n_prev, sp_prev, phi_prev, lambda_prev, pi_prev, A_prev, 
                              params, rho, Zcounts, 
                              estimateNormal = True, estimatePloidy = True,
                              estimatePrecision = True, estimateInitDist = True, 
                              estimateTransition = True, estimateSubclone = True,
                              verbose = True):
    """


    Parameters
    ----------

    
    Returns
    -------


    """

    KS = rho.shape[0]
    K = len(params["ct"])
    T = rho.shape[1]
    S = len(n_prev)

    intervalNormal = (1e-6, 1 - 1e-6)
    intervalSubclone = (1e-6, 1 - 1e-6)
    intervalPhi = (np.finfo(np.double).eps, 10)
    intervalLambda = (1e-10, 1e4)

    # initialize params to be estimated
    n_hat = n_prev
    sp_hat = sp_prev
    phi_hat = phi_prev
    lambda_hat = lambda_prev
    pi_hat = pi_prev
    A_hat = A_prev

    # Update transition matrix A
    if estimateTransition:
        for k in range(KS):
            A_hat[k,:] = Zcounts[k,:] + params["dirPrior"][k,:]
            A_hat[k,:] = hmm_utilities.normalize(A_hat[k,:])

    # map estimate for pi (initial state dist)
    if estimateInitDist:
        pi_hat = estimateMixWeightsParamMap(rho, params["kappa"])

    # map estimate for normal 
    if estimateNormal:
        estNorm = minimize(completeLikelihoodFun, n_prev, args=(np.repeat("n", S), n_prev, sp_prev, phi_prev, 
                                                        lambda_prev, pi_hat, A_hat, params, D, rho, Zcounts, 
                                                        estimateNormal, estimatePloidy, estimatePrecision, estimateInitDist,
                                                        estimateTransition, estimateSubclone), method='L-BFGS-B',
                                                        bounds=Bounds(intervalNormal[0],intervalNormal[1]),
                                                        options={'eps':1e-8})
        n_hat = estNorm.x

    if estimateSubclone:
        estSubclone = minimize(completeLikelihoodFun, sp_prev, args=(np.repeat("sp", S), n_hat, sp_prev, phi_prev, 
                                                        lambda_prev, pi_hat, A_hat, params, D, rho, Zcounts, 
                                                        estimateNormal, estimatePloidy, estimatePrecision, estimateInitDist,
                                                        estimateTransition, estimateSubclone), method='L-BFGS-B',
                                                        bounds=Bounds(intervalNormal[0],intervalNormal[1]),
                                                        options={'eps':1e-8})
        sp_hat = estSubclone.x
    
    if estimatePloidy:
        estPhi = minimize(completeLikelihoodFun, phi_prev, args=(np.repeat("phi", len(phi_prev)), n_hat, sp_hat, phi_prev, 
                                                        lambda_prev, pi_hat, A_hat, params, D, rho, Zcounts, 
                                                        estimateNormal, estimatePloidy, estimatePrecision, estimateInitDist,
                                                        estimateTransition, estimateSubclone), method='L-BFGS-B',
                                                        bounds=Bounds(intervalPhi[0],intervalPhi[1]),
                                                        options={'eps':1e-8})
        phi_hat = estPhi.x

    if estimatePrecision:
        estLambda = minimize(completeLikelihoodFun, lambda_prev, args=(np.repeat("lambda", K*S), n_hat, sp_hat, phi_hat, 
                                                        lambda_prev, pi_hat, A_hat, params, D, rho, Zcounts, 
                                                        estimateNormal, estimatePloidy, estimatePrecision, estimateInitDist,
                                                        estimateTransition, estimateSubclone), method='L-BFGS-B',
                                                        bounds=Bounds(intervalLambda[0],np.inf),
                                                        options={'eps':1e-8})
        lambda_hat = estLambda.x

    return {"n":n_hat, "sp":sp_hat, "phi":phi_hat, "lambda":lambda_hat, "piG":pi_hat, "A":A_hat, "F":estLambda.fun*-1}



def runEM(copy, chroms, chrTrain, param, maxiter, verbose=False, 
            estimateNormal=True, estimatePloidy=True, estimatePrecision=True,
            estimateTransition=True, estimateInitDist=True, estimateSubclone=True,
            likChangeConvergence=1e-3):
    """


    Parameters
    ----------
        copy : pandas.DataFrame
    
    Returns
    -------


    """

    if copy.shape[0] != len(chroms) or copy.shape[0] != len(chrTrain):
        raise IndexError("runEM: Length of inputs do not match for one of: copy, chr, chrTrain")

    if param["ct"] is None or param["lambda"] is None or param["nu"] is None or param["kappa"] is None:
        raise IndexError("runEM: Parameter missing, ensure all parameters exist as columns in data frame: ct, lambda, nu, kappa")

    S = param["numberSamples"]
    K = len(param["ct"])
    Z = np.sum(param["ct_sc_status"]) #number of subclonal states (# repeated copy number states)
    KS = K**S
    N = copy.shape[0]
    rho = np.zeros((KS, N))
    py = np.zeros((KS, N), order="F")            # Local evidence
    mus = [np.zeros((KS, S)) for i in range(maxiter)]     # State means
    lambdas = [np.zeros((K, S)) for i in range(maxiter)]  # State Variances
    phi = np.tile(np.nan, (len(param["phi_0"]), maxiter))					 # Tumour ploidy
    n = np.tile(np.nan, (S, maxiter))						 # Normal contamination
    sp = np.tile(np.nan, (S, maxiter))     # cellular prevalence (tumor does not contain event)
    piG = np.zeros((KS, maxiter), order="F")     # Initial state distribution
    converged = False               # Flag for convergence
    Zcounts = np.zeros((KS,KS))
    loglik = np.zeros(maxiter)

    ptmTotal = time.perf_counter() # start total timer

    # SET UP
    # Set up the chromosome indices and make cell array of chromosome indicies
    indexes = np.unique(chroms, return_index=True)[1]
    chrs = np.array([chroms[index] for index in sorted(indexes)])
    chrsI = []

    # initialise the chromosome index and the init state distributions
    for i in range(len(chrs)):
        chrsI.append(np.where(chroms == chrs[i])[0])
    
    # INITIALIZATION
    if verbose: print("runEM: Initialization")
    i = 0
    piG[:, i] = hmm_utilities.normalize(param["kappa"])
    n[:, i] = param["n_0"]
    sp[:, i] = param["sp_0"]
    phi[:, i] = param["phi_0"]
    lambdas[i] = param["lambda"]
    lambdasKS = pd.DataFrame(lambdas[0])
    mus[i] = get2and3ComponentMixture(param["jointCNstates"].values, param["jointSCstatus"].values, n[:, i], sp[:, i], phi[:, i])

    # Likelihood #
    for ks in range(KS):
        #probs = tdistPDF(copy.values, mus[i][ks,:], lambdasKS.iloc[ks,:], param["nu"])
        probs = tdistPDF(copy.values, mus[i][ks,:], lambdasKS.values[ks,:], param["nu"])
        if probs.ndim > 1:
            py[ks,:] = probs.apply(prod, axis=0) # multiply across samples for each data point to get joint likelihood.
        else:
            py[ks,:] = probs

    # initialize transition prior #
    A = hmm_utilities.normalize(param["A"])
    A_prior = A
    dirPrior = param["dirPrior"]

    loglik[i] = -np.inf

    while converged == False and i < maxiter-1:
        ptm = time.perf_counter()

        i = i + 1

        ################ E-step ####################
        if verbose: print("runEM iter", i-1 , ": Expectation")

        Zcounts = np.zeros((KS, KS))
        for j in range(len(chrsI)):
            I = chrsI[j][np.in1d(chrsI[j], np.where(chrTrain)[0])]
            if len(I) > 0:
                #output = forward_backward.forward_backward(piG[:, i - 1], A, py[:, I])
                output = forward_backward(piG[:, i - 1], A, py[:, I])
                rho[:, I] = output["rho"]
                loglik[i] = loglik[i] + output["loglik"]
                Zcounts = Zcounts + output["xi"].sum(axis=2)
            
        ################ M-step ####################
        if verbose: print("runEM iter", i-1 , ": Maximization")

        output = estimateParamsMap(copy.loc[chrTrain, :], n[:, i - 1], sp[:, i - 1], phi[:, i - 1], 
                                lambdas[i - 1], piG[:, i - 1], A, param, rho[:, chrTrain], Zcounts, 
                                estimateNormal = estimateNormal, estimatePloidy = estimatePloidy,
                                estimatePrecision = estimatePrecision, estimateTransition = estimateTransition,
                                estimateInitDist = estimateInitDist, estimateSubclone = estimateSubclone)

        if verbose :
            for s in range(S):
                print("Sample", s, " n=", output["n"][s], ", sp=", output["sp"][s], 
                ", phi=", output["phi"][s], 
                ", lambda=", output["lambda"][s],
                ", F=", output["F"])
        
        n[:, i] = output["n"]
        sp[:, i] = output["sp"]
        phi[:, i] = output["phi"]
        lambdas[i] = output["lambda"]
        piG[:, i] = output["piG"]
        A = output["A"]
        estF = output["F"]

        # Recalculate the likelihood
        lambdasKS = pd.DataFrame(lambdas[i])
        mus[i] = get2and3ComponentMixture(param["jointCNstates"].values, param["jointSCstatus"].values, n[:, i], sp[:, i], phi[:, i])
        for ks in range(KS):
            #probs = tdistPDF(copy.values, mus[i][ks,:], lambdasKS.iloc[ks,:], param["nu"])
            probs = tdistPDF(copy.values, mus[i][ks,:], lambdasKS.values[ks,:], param["nu"])
            if probs.ndim > 1:
                py[ks,:] = probs.apply(prod, axis=0) # multiply across samples for each data point to get joint likelihood.
            else:
                py[ks,:] = probs

        prior = priorProbs(n[:, i], sp[:, i], phi[:, i], lambdas[i], piG[:, i], A, param, 
                        estimateNormal = estimateNormal, estimatePloidy = estimatePloidy,
                        estimatePrecision = estimatePrecision, estimateTransition = estimateTransition,
                        estimateInitDist = estimateInitDist, estimateSubclone = estimateSubclone)

        # check converence 
        loglik[i] = loglik[i] + prior["prior"]
        elapsedTime = time.perf_counter() - ptm
        if verbose:
            print("runEM iter", i-1, " Log likelihood:", loglik[i]) 
            print("runEM iter", i-1, " Time: ", elapsedTime / 60, " min.")

        if (np.abs(loglik[i] - loglik[i - 1]) / np.abs(loglik[i])) < likChangeConvergence:
            converged = True

        if loglik[i] < loglik[i - 1]:
            i = i - 1
            converged = True
        
    if converged:
        # Perform one last round of E-step to get latest responsibilities
        # E-step
        if verbose: print("runEM iter", i-1 ,": Re-calculating responsibilties from converged parameters.")
        for j in range(len(chrsI)):
            I = chrsI[j]
            output = forward_backward(piG[:, i], A, py[:, I])
            rho[:, I] = output["rho"]
    
    if verbose:
        totalTime = time.perf_counter() - ptmTotal
        print("runEM: Using optimal parameters from iter", i-1)
        print("runEM: Total elapsed time: ", totalTime / 60, "min.")
    
    #### Return parameters ####
    n = n[:, :i+1]
    sp = sp[:, :i+1]
    phi = phi[:, :i+1]
    mus = mus[:i+1]
    lambdas = lambdas[:i+1]
    piG = piG[:, :i+1]
    loglik = loglik[:i+1]

    output = {}
    output["n"] = n
    output["sp"] = sp
    output["phi"] = phi
    output["mus"] = mus
    output["lambdas"] = lambdas
    output["pi"] = piG
    output["A"] = A
    output["loglik"] = loglik
    output["rho"] = rho
    output["param"] = param
    output["py"] = py
    output["iter"] = i

    return output